# Data loading functionalities
import polars as pl
from .seq import Assembly
from .feature import Pileup

def load_fasta(path, trim_names=False, trim_character=" ") -> dict:
    """
    Reads a fasta file and returns a dictionary with the contig names as 
    keys and the sequences as values
    """
    with open(path, 'r') as f:
        lines = f.readlines()
    data = {}
    active_sequence_name = "no_header"
    for line in lines:
        line = line.rstrip()
        if line[0] == '>':
            active_sequence_name = line[1:]
            if trim_names:
                active_sequence_name = active_sequence_name.split(trim_character)[0]
            if active_sequence_name not in data:
                data[active_sequence_name] = ''
        else:
            data[active_sequence_name] += line
    return data

def load_pileup(path: str, min_coverage: int, min_fraction: float = 0):
    """
    Load pileup file from path to pileup.bed output of modkit pileup
    """
    pileup = (
        pl.scan_csv(path, separator = "\t", has_header = False)
        .filter(pl.col("column_10") > min_coverage)
        .filter(pl.col("column_11") > min_fraction*100)
        .filter(pl.col("column_10") / (pl.col("column_10") + pl.col("column_17")) > 0.3)
        .select(["column_1", "column_2","column_4", "column_6", "column_11", "column_10"])
        .with_columns(pl.col("column_11") / 100)
        .collect()
    )
    pileup = pileup.rename({"column_1":"contig", "column_2": "position", "column_4": "mod_type", "column_6": "strand", "column_11": "fraction_mod", "column_10":"Nvalid_cov"})
    return Pileup(pileup)

def extract_contig_mods_with_sufficient_information(pileup: Pileup, assembly: Assembly, min_mods_pr_contig: int, min_mod_frequency: int):
    contigs_in_assembly = list(assembly.assembly.keys())
    pileup = pileup.filter(pl.col("contig").is_in(contigs_in_assembly))

    contigmods_with_more_than_min_mods = pileup.group_by("contig_mod").count().filter(
            pl.col("count") > min_mods_pr_contig
        ).get_column("contig_mod").to_list()

    assm_lengths = pl.DataFrame({
        "contig": list(assembly.assembly.keys()),
        "length": [len(contig) for contig in assembly.assembly.values()]
    })
    contigmods_with_sufficient_mod_frequency = pileup \
        .group_by(["contig", "mod_type"]) \
        .agg(pl.count()) \
        .join(assm_lengths, on = "contig") \
        .filter(pl.col("count") > pl.col("length")/min_mod_frequency) \
        .with_columns([
            (pl.col("contig") + "_" + pl.col("mod_type")).alias("contig_mod")
        ]) \
        .get_column("contig_mod").unique().to_list()
    
    # Get contig_mods to keep and to remove
    contig_mods_to_keep = list(set(contigmods_with_more_than_min_mods) & set(contigmods_with_sufficient_mod_frequency))
    contig_mods_to_remove = list(set(pileup.get_column("contig_mod").unique().to_list()) - set(contig_mods_to_keep))
    return contig_mods_to_keep, contig_mods_to_remove


def load_low_coverage_positions(path_pileup: str, contig_mods_to_load: list[str], min_coverage: float = 5):
    """
    Load pileup file from path to pileup.bed output of modkit pileup
    """
    pileup = (
        pl.scan_csv(path_pileup, separator = "\t", has_header = False)
        .filter(pl.col("column_10") <= min_coverage)
        .filter(pl.col("column_10") / (pl.col("column_10") + pl.col("column_17")) > 0.3)
        .select(["column_1", "column_2","column_4", "column_6", "column_10"])
        .with_columns((pl.col("column_1") + "_" + pl.col("column_4")).alias("contig_mod"))
        .filter(pl.col("contig_mod").is_in(contig_mods_to_load))
        .group_by("column_1", "column_4")
        .agg(
            pl.col("column_2"),
            pl.col("column_6")
        )
        .collect()
    )
    pileup = pileup.rename({"column_1":"contig", "column_2": "position", "column_4": "mod_type", "column_6": "strand"})
    return pileup


def load_assembly(path: str):
    """
    Load assembly from path to fasta file
    """
    return Assembly(load_fasta(path))



import hdbscan
from sklearn.cluster import AgglomerativeClustering, SpectralClustering
from sklearn.mixture import GaussianMixture
from sklearn.decomposition import PCA
import polars as pl
from nanomotif.binnary import data_processing as dp
import logging

def detect_contamination(contig_methylation, contig_lengths, num_consensus = 4, threads = 1, spectral_n_neighbors=10):
    logger = logging.getLogger(__name__)
    logger.info("Starting contamination detection analysis...")

    # Create the bin_consensus dataframe for scoring
    logger.info("Creating bin_consensus dataframe for scoring...")
    contig_methylation = dp.impute_contig_methylation_within_bin(
        contig_methylation
    )
    # INFO: It does not makes sense to find contamination in bins with a single contig
    single_contig_bins = contig_methylation\
        .select(["bin", "contig"])\
        .unique()\
        .group_by("bin")\
        .agg(pl.col("contig").count().alias("n_contigs"))\
        .filter(pl.col("n_contigs") == 1)\
        .get_column("bin")

    contig_methylation = contig_methylation\
        .filter(~pl.col("bin").is_in(single_contig_bins))

    contig_names, matrix = dp.create_matrix(contig_methylation)

    original_features = matrix.shape[1]
    print(f"Original number of features: {original_features}")

    print("Applying PCA for feature decorrelation")
    pca_variance=0.90
    pca = PCA(n_components=pca_variance, svd_solver = "full")
    matrix = pca.fit_transform(matrix)

    reduced_components = matrix.shape[1]
    print(f"PCA reduced the feature space from {original_features} to {reduced_components}")
    print(f"Total explained variance by pca: {pca.explained_variance_ratio_.sum():.2f}")
    
    n_bins = len(contig_methylation.get_column("bin").unique())

    print("Running Spectral")
    spectral = SpectralClustering(n_clusters=n_bins, affinity = 'nearest_neighbors', random_state=42, n_jobs = threads, n_neighbors = spectral_n_neighbors)
    spectral_labels = spectral.fit_predict(matrix)

    print("Running Agglomerative Clustering")
    agg = AgglomerativeClustering(n_clusters = n_bins)
    agg_labels = agg.fit(matrix).labels_
    
    print("Running HDBSCAN")
    dscan = hdbscan.HDBSCAN(min_samples=3, min_cluster_size=2, metric = "euclidean", allow_single_cluster=False)
    dscan_labels = dscan.fit_predict(matrix)

    print("Running Gaussian Mixture Model")
    gmm = GaussianMixture(n_components=n_bins, covariance_type='full', random_state=42)
    gmm.fit(matrix)
    gmm_labels = gmm.predict(matrix)


    contig_bin = contig_methylation\
        .select(["bin", "contig"])\
        .unique() 

    results = pl.DataFrame({
                    "contig": contig_names,
                    "spectral": spectral_labels,
                    "agg": agg_labels,
                    "hdbscan": dscan_labels,
                    "gmm": gmm_labels
                })\
                .join(contig_bin, on="contig")\
                .melt(
                    id_vars = ["contig", "bin"],
                    value_vars=["spectral", "agg", "hdbscan", "gmm"],
                    value_name="cluster",
                    variable_name="method"
                )


    bin_size = contig_bin\
        .join(contig_lengths, on = "contig")\
        .group_by("bin")\
        .agg(
            pl.col("length").sum().alias("bin_length"),
            pl.col("contig").count().alias("n_contigs_bin")
        )

    cluster_sizes = results\
        .join(contig_lengths, on="contig")\
        .group_by(["bin","method", "cluster"])\
        .agg(
            pl.col("contig").count().alias("n_contigs"),
            pl.col("length").sum().alias("cluster_length")
        )\
        .join(bin_size, on = "bin")\
        .with_columns(
            (pl.col("n_contigs") / pl.col("n_contigs_bin")).alias("fraction_contigs"),
            (pl.col("cluster_length") / pl.col("bin_length")).alias("fraction_length")
        )

    assigned_cluster = cluster_sizes\
        .group_by(["method","bin"])\
        .agg(
            pl.col("cluster_length").max().alias("cluster_length")
        )\
        .join(cluster_sizes, on = ["bin", "method","cluster_length"], how = "left")\
        .rename({
                    "cluster": "bin_cluster"
                })\
        .drop(["cluster_length", "n_contigs"])\
        .filter(pl.col("fraction_length") >= 0.85)



    results = results\
        .join(assigned_cluster, on = ["bin", "method"])\
        .sort(["bin", "contig"])


    confident_contaminants = results\
        .with_columns(
           (pl.col("bin_cluster") != pl.col("cluster")).cast(pl.Int8).alias("is_contaminant")
        )\
        .group_by(["contig"])\
        .agg(
            pl.col("is_contaminant").sum().alias("sum_predictions"),
        )\
        .filter(pl.col("sum_predictions") >= num_consensus)\
        .get_column("contig")

    contamination_contigs = results\
        .filter(pl.col("contig").is_in(confident_contaminants))
    
    return contamination_contigs



import polars as pl
import pandas as pd
import numpy as np
np.random.seed(1)

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

import gzip
import os
import logging

def load_data(args):
    """
    Load data for the analysis.
    """
    logger = logging.getLogger(__name__)
    logger.info("Loading data.")
    # Load the data from the provided files
    # Load the data from the provided files using Polars
    bin_motifs = pl.read_csv(args.bin_motifs, separator="\t")
    
    # Polars automatically infers headers; if args.contig_bins file doesn't have headers, you need to specify it
    contig_bins = pl.read_csv(args.contig_bins, separator="\t", has_header=False, infer_schema_length=10000) \
        .rename({
            "column_1": "contig",
            "column_2": "bin"
        })

    logger.info("Data loaded.")
    return bin_motifs, contig_bins

def read_fasta(file_path):
    # Check if the file exists
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")    
    
    # Check if the file has a valid FASTA extension
    valid_extensions = ['.fasta', '.fa', '.fna', '.gz']
    if not any(file_path.endswith(ext) for ext in valid_extensions):
        raise ValueError(f"Unsupported file extension. Please provide a FASTA file with one of the following extensions: {', '.join(valid_extensions)}")

    # Check if the file is a gzipped FASTA file
    if file_path.endswith('.gz'):
        with gzip.open(file_path, "rt") as handle:  # "rt" mode for reading as text
            return {record.id: str(record.seq) for record in SeqIO.parse(handle, "fasta")}
    else:
        # Read a regular (uncompressed) FASTA file
        with open(file_path, "r") as handle:
            return {record.id: str(record.seq) for record in SeqIO.parse(handle, "fasta")}

def find_contig_lengths(assembly):
    # Parse the assembly file using Biopython
    contigs = []
    lengths = []
    
    for id, seq in assembly.items():
        contigs.append(id)       # Contig name (FASTA header)
        lengths.append(len(seq)) # Contig length
    
    # Create a Polars DataFrame
    df = pl.DataFrame({
        "contig": contigs,
        "length": lengths
    })
    
    return df

def write_bins_from_contigs(new_contig_bins, assembly_dict, output_dir):
    """
    Creates new bin files based on contig assignments in new_contig_bins DataFrame.

    Args:
    - new_contig_bins (DataFrame): DataFrame with contig and bin assignments.
    - assembly_dict (dict): Dictionary with contig IDs as keys and sequences as values, from the assembly file.
    - output_dir (str): Path to the directory where the new bin files will be saved.
    """
    logger = logging.getLogger(__name__)
    
    # check if the output directory exists
    logger.info(f"Writing bins to {output_dir}")
    output_bins_dir = os.path.join(output_dir)
    if not os.path.exists(output_bins_dir):
        os.makedirs(output_bins_dir)
    
    # Group the DataFrame by 'bin'
    grouped = new_contig_bins.groupby('bin')

    for bin_name, group in grouped:
        # Initialize a list to hold SeqRecord objects for the current bin
        bin_records = []

        for contig in group['contig']:
            # Create a SeqRecord object for each contig in the bin, if it exists in the assembly dictionary
            if contig in assembly_dict:
                seq_record = SeqRecord(Seq(assembly_dict[contig]), id=contig, description="")
                bin_records.append(seq_record)

        # Define the output file name for the current bin
        output_file = f"{bin_name}.fa"
        output_path = os.path.join(output_bins_dir, output_file)

        # Write the SeqRecord objects to a FASTA file
        with open(output_path, "w") as output_handle:
            SeqIO.write(bin_records, output_handle, "fasta")

        logger.info(f"Written {len(bin_records)} contigs to {output_file}")



def generate_output(output_df, outdir, filename, header=True):
    """
    Generate the output files for the analysis.
    """
    # If the output directory does not exist, create it
    if not os.path.exists(outdir) and outdir != "":
        os.makedirs(outdir)
    
    file_path = os.path.join(outdir, filename)
    
    # Generate the output files
    output_df.to_csv(file_path, sep="\t", index=False, header=header)


def prepare_bin_consensus(bin_motifs, args):
    """
    Prepares the bin_consensus_from_bin_motifs DataFrame by calculating the mean methylation per bin and motif_mod and converting it to binary.    
    """
    # Combine 'motif' and 'mod_type' into 'motif_mod'
    bin_motifs = bin_motifs.with_columns((pl.col("motif") + "_" + pl.col("mod_type") + "_" + pl.col("mod_position").cast(pl.Utf8)).alias("motif_mod"))

    # Calculate total motifs and mean methylation
    bin_motifs = bin_motifs.with_columns([
        (pl.col("n_mod_bin") + pl.col("n_nomod_bin")).alias("n_motifs"),
        (pl.col("n_mod_bin") / (pl.col("n_mod_bin") + pl.col("n_nomod_bin"))).alias("mean_methylation")
    ])

    # Create 'methylation_binary' column based on mean methylation cutoff
    bin_motifs = bin_motifs.with_columns((pl.col("mean_methylation") >= args.mean_methylation_cutoff).cast(int).alias("methylation_binary"))

    # Filter motifs that are not observed more than n_motif_bin_cutoff times
    bin_motifs = bin_motifs.filter(pl.col("n_motifs") >= args.n_motif_bin_cutoff)

    # Filter for rows where 'methylation_binary' is 1 and select relevant columns
    bin_motif_binary = bin_motifs.filter(pl.col("methylation_binary") == 1)[["bin", "motif_mod", "mean_methylation", "methylation_binary"]]

    # Create a binary index for bin and motif_mod
    # Assuming unique_bins and unique_motif_mods are lists of unique values
    unique_bins = bin_motif_binary["bin"].unique().to_list()
    unique_motif_mods = bin_motif_binary["motif_mod"].unique().to_list()

    # Create a DataFrame with all combinations of "bin" and "motif_mod"
    bin_motif_combinations = pl.DataFrame({
        "bin": [bin for bin in unique_bins for _ in unique_motif_mods],
        "motif_mod": [motif_mod for _ in unique_bins for motif_mod in unique_motif_mods]
    })

    # Join the combinations DataFrame with bin_motif_binary
    bin_motif_binary_enriched = bin_motif_combinations.join(
        bin_motif_binary, on=["bin", "motif_mod"], how="left"
    )

    # Fill missing values if necessary and adjust columns as needed
    bin_motif_binary_enriched = bin_motif_binary_enriched.with_columns([
        pl.col("mean_methylation").fill_null(0).alias("mean_methylation"),
        pl.col("methylation_binary").fill_null(0).alias("methylation_binary")
    ])
    

    return bin_motif_binary_enriched


def add_bin(contig_methylation, contig_bins):
    # Filter and enhance contig_methylation 
    contig_methylation_w_bin = contig_methylation \
        .with_columns(
            (pl.col("motif") + "_" + pl.col("mod_type") + "_" + pl.col("mod_position").cast(pl.Utf8)).alias("motif_mod")
        )
    
    # Merge with contig_bins
    contig_methylation_w_bin = contig_methylation_w_bin.join(contig_bins, on="contig", how="left") \
        .with_columns(pl.col("bin").fill_null("unbinned")) \
        .drop(["mod_position", "mod_type", "motif"])
    
    return contig_methylation_w_bin


def impute_contig_methylation_within_bin(contig_methylation):
    contig_methylation = contig_methylation\
        .filter(pl.col("bin") != "unbinned")
    
    bin_methylation = contig_methylation \
        .group_by(["bin", "motif_mod"]) \
        .agg(
            (
                (pl.col("median") * pl.col("N_motif_obs")).sum() / pl.col("N_motif_obs").sum()
            ).alias("mean_bin_median")
        )

    contig_bin_cross = contig_methylation\
        .unique(subset=["contig", "bin"], maintain_order = True)\
        .select(["contig", "bin"])

    contig_methylation_imputed = contig_bin_cross\
        .join(bin_methylation, on = ["bin"], how = "left")\
        .join(contig_methylation, on = ["bin", "contig", "motif_mod"], how = "left")\
        .sort(["bin", "contig", "motif_mod"])\
        .with_columns(
            pl.when(pl.col("median").is_null()).then(pl.col("mean_bin_median")).otherwise(pl.col("median")).alias("median")
        )\
        .drop("N_motif_obs")
        
    return contig_methylation_imputed

def impute_unbinned_contigs(contig_methylation):
    unbinned_contig_methylation = contig_methylation\
        .filter(pl.col("bin") == "unbinned")

    unbinned_contigs = unbinned_contig_methylation\
        .select(["contig"])\
        .unique()
        
    cross_motif_mod_contig = contig_methylation\
        .select(["motif_mod"])\
        .unique()\
        .join(unbinned_contigs, how = "cross")\
        .with_columns(
            pl.lit("unbinned").alias("bin")
        )\
        .sort(["contig", "motif_mod"])

    num_rows = cross_motif_mod_contig.height
    random_values = np.random.uniform(0.0, 0.15, size = num_rows)
    
    cross_motif_mod_contig = cross_motif_mod_contig\
        .with_columns(
            pl.Series("pseudo_median", random_values)
        )
    
    imputed_unbinned_contig_methylation = cross_motif_mod_contig\
        .join(unbinned_contig_methylation, on=["contig", "motif_mod", "bin"], how="left")\
        .with_columns(
            pl.when(pl.col("median").is_null())
                .then(pl.col("pseudo_median"))
                .otherwise(pl.col("median"))
                .alias("median")
        )\
        .drop("pseudo_median")\
        .sort(["contig", "motif_mod"])

    return imputed_unbinned_contig_methylation
    

def create_matrix(contig_methylation):
    matrix_df = contig_methylation\
        .select(["contig", "motif_mod", "median"])\
        .pivot(
            values = "median",
            index = "contig",
            columns="motif_mod",
            aggregate_function = None,
            sort_columns = True
        )\
        .fill_null(0)

    contig_names = matrix_df.get_column("contig")
    matrix = matrix_df.drop("contig").to_numpy()
    return contig_names, matrix

    
def load_contamination_file(contamination_file):
    """
    Load the contamination file from the provided path.
    """
    contamination = pl.read_csv(contamination_file, separator= "\t")
    
    # Check if the file contains the required columns
    required_columns = ["contig"]
    if not all(column in contamination.columns for column in required_columns):
        raise ValueError("The contamination file does not contain the required columns.")
    
    return contamination


def create_contig_bin_file(contig_bins, contamination, include=None):
    """
    Create a new contig_bin file based on the analysis results and contamination file.
    """
    # Remove contigs in the contamination file from the contig_bins
    contig_bins = contig_bins[~contig_bins["contig"].isin(contamination["contig"])]
    
    # Add the contigs in the include DataFrame to the contig_bins
    if include is not None:
        contig_bins = pd.concat([contig_bins, include[["contig", "bin"]]], ignore_index=True)
    
    # Sort the contig_bins by bin and contig
    contig_bins = contig_bins.sort_values(by=["bin", "contig"])
    
    return contig_bins



import argparse
from nanomotif._version import __version__
import os


def create_parser():
    formatter = lambda prog: argparse.HelpFormatter(prog,max_help_position=28)
    parser = argparse.ArgumentParser(description="Motif identification and utilisation commands", formatter_class=formatter)
    parser.add_argument("--version", action="version", version="%(prog)s {}".format(__version__))
    subparsers = parser.add_subparsers(help="-- Command descriptions --", dest="command")

    ###########################################################################
    # Shared sub-command arguments
    parser_positional = argparse.ArgumentParser(add_help=False)
    parser_positional.add_argument("assembly", type=str, help="path to the assembly file.")
    parser_positional.add_argument("pileup", type=str, help="path to the modkit pileup file.")

    parser_optional = argparse.ArgumentParser(add_help=False)
    parser_optional.add_argument("--out", type=str, help="path to the output folder", default="nanomotif")
    parser_optional.add_argument("-t", "--threads", type=int, default=1, help="number of threads to use. Default is 1")
    parser_optional.add_argument("-v", "--verbose", action="store_true", help="increase output verbosity. (set logger to debug level)")
    parser_optional.add_argument("--seed", type=int, default=1, help="seed for random number generator. Default: %(default)s")
    parser_optional.add_argument("--threshold_methylation_general", type=float, default=0.70, help="minimum fraction of reads that must be methylated at a position for the position to be methylated. These position are used for counting number of methylated position of a motif. Default: %(default)s")

    ###########################################################################
    # Find Motifs
    parser_shared_find_motifs = argparse.ArgumentParser(add_help=False)
    parser_shared_find_motifs.add_argument("--search_frame_size", type=int, default=40, help="length of the sequnces sampled around confident methylatyion sites. Default: %(default)s")
    parser_shared_find_motifs.add_argument("--threshold_methylation_confident", type=float, default=0.80, help="minimum fraction of reads that must be methylated at a position for the position to be considered confiently methylated. These position are used to search for candidate motifs. Default: %(default)s")
    parser_shared_find_motifs.add_argument("--threshold_valid_coverage", type=int, default=5, help="minimum valid base coverage for a position to be considered. Default: %(default)s")
    parser_shared_find_motifs.add_argument("--minimum_kl_divergence", type=float, default=0.05, help="minimum KL-divergence for a position to considered for expansion in  motif search. Higher value means less exhaustive, but faster search. Default: %(default)s")
    parser_shared_find_motifs.add_argument("--min_motifs_contig", type=int, default=20, help="minimum number of times a motif has to have been oberserved in a contig. Default: %(default)s")
    parser_shared_find_motifs.add_argument("--read_level_methylation", action="store_true", help="If specified, methylation is calculated on read level instead of contig level. This is slower but produces more stable motifs.")
    parser_shared_find_motifs.add_argument("--min_motif_score", type=float, default=0.2, help="minimum score for a motif to be kept after identification considered valid. Default: %(default)s")
    parser_find_motifs = subparsers.add_parser(
        'find_motifs', 
        parents=[parser_positional, parser_optional, parser_shared_find_motifs], 
        help="Finds motifs directly on contig level in provided assembly"
    )

    ###########################################################################
    # Score motifs
    parser_score_motifs = subparsers.add_parser(
        'score_motifs', 
        parents=[parser_positional, parser_optional],
        help="Find motifs indirectly in contigs by scoring with motifs found in other contigs"
    )
    parser_score_motifs.add_argument("motifs", type=str, help="path to the motifs file.")
    parser_score_motifs.add_argument("--save-motif-positions", action="store_true", help="save motif positions in the output folder")
    parser_score_motifs.add_argument("--threshold_valid_coverage", type=int, default=5, help="minimum valid base coverage for a position to be considered. Default: %(default)s")
    
    ###########################################################################
    # Bin consensus
    parser_shared_bin_consensus = argparse.ArgumentParser(add_help=False, conflict_handler="resolve")
    parser_shared_bin_consensus.add_argument("--min_motifs_bin", type=int, default=50, help="minimum number of times a motif has to have been oberserved in a bin. Default: %(default)s")
    parser_bin_consensus = subparsers.add_parser(
        'bin_consensus', 
        parents=[parser_positional, parser_optional, parser_shared_bin_consensus],
        help="Indentifies highly methylated motifs in bins"
    )
    parser_bin_consensus.add_argument("motifs", type=str, help="path to the motifs file.")
    parser_bin_consensus.add_argument("bins", type=str, help="tsv file specifying which bin contigs belong.")
    parser_bin_consensus.add_argument("motifs_scored", metavar="motifs-scored", type=str, help="path to the motif-scored file.")
    parser_bin_consensus.add_argument("--threshold_valid_coverage", type=int, default=5, help="minimum valid base coverage for a position to be considered. Default: %(default)s")

    ###########################################################################
    # Complete workflow
    parser_complete_workflow = subparsers.add_parser('motif_discovery', help='Runs find_motifs, score_motifs and bin_consensus', parents=[parser_positional, parser_optional, parser_shared_find_motifs, parser_shared_bin_consensus], conflict_handler="resolve")
    parser_complete_workflow.add_argument("bins", type=str, help="tsv file specifying which bin contigs belong.")

    parser_complete_workflow.add_argument("--save-motif-positions", action="store_true", help="save motif positions in the output folder")
    
    ###########################################################################
    # Bin contamination and inclusion   
    parser_binnary_shared = argparse.ArgumentParser(description="Contamination DNA Methylation Pattern", add_help=False)
    """Function to add common arguments to a subparser."""
    
    parser_binnary_shared_mandatory = parser_binnary_shared.add_argument_group("Mandatory Arguments")
    
    
    parser_binnary_shared_mandatory.add_argument(
        "--pileup", type=str, help="Path to pileup.bed", required=True
    )
    parser_binnary_shared_mandatory.add_argument(
        "--assembly", type=str, help="Path to assembly file [fasta format required]", required=True
    )
    parser_binnary_shared_mandatory.add_argument("--bin_motifs", type=str, help="Path to bin-motifs.tsv file", required=True)
    parser_binnary_shared_mandatory.add_argument(
        "--contig_bins", type=str, help="Path to bins.tsv file for contig bins", required=True
    )
    parser_binnary_shared.add_argument("-t", "--threads", type=int, default=1, help="Number of threads to use for multiprocessing")

    parser_binnary_shared.add_argument(
        "--min_valid_read_coverage",
        type=int,
        default=3,
        help="Minimum read coverage for calculating methylation [used with methylation_util executable]",
    )
    parser_binnary_shared.add_argument(
        "--methylation_threshold",
        type=int,
        default=24,
        help="Filtering criteria for trusting contig methylation. It is the product of mean_read_coverage and N_motif_observation. Higher value means stricter criteria. [default: 24]",
    )
    
    parser_binnary_shared.add_argument(
        "--num_consensus",
        type=int,
        default=4,
        help="Number of models that has to agree for classifying as contaminant",
    )
    parser_binnary_shared.add_argument(
        "--force",
        action='store_true',
        help="Force override of motifs-scored-read-methylation.tsv. If not set existing file will be used.",
    )
    
    parser_binnary_shared.add_argument(
        "--write_bins",
        action='store_true',
        help="If specified, new bins will be written to a bins folder. Requires --assembly_file to be specified.",
    )
    
    parser_binnary_shared_mandatory.add_argument("--out", type=str, help="Path to output directory", required=True, default="nanomotif")
    
    # Binnary contamination
    parser_contamination = subparsers.add_parser(
        'detect_contamination', 
        help="Detect contamination in bins",
        parents=[parser_binnary_shared]
    )  
    
    parser_contamination.add_argument(
        '--contamination_file',
        type=str,
        help="Path to an existing contamination file if bins should be outputtet as a post-processing step"
    )
    
    # Binnary inclusion
    parser_inclusion = subparsers.add_parser(
        'include_contigs', 
        help="Include contigs in bins",
        parents=[parser_binnary_shared]
    )
    parser_inclusion.add_argument(
        "--mean_model_confidence",
        type = float,
        help = "Mean probability between models for including contig. Contigs above this value will be included. [default: 0.8]",
        default = 0.8
    )
    
    group = parser_inclusion.add_mutually_exclusive_group(required=False)
    group.add_argument(
        "--contamination_file",
        type=str,
        help="Path to an existing contamination file to include in the analysis"
    )
    group.add_argument(
        "--run_detect_contamination",
        action='store_true',
        help="Indicate that the detect_contamination workflow should be run first"
    )
    
    ###########################################################################
    # MTase-linker
    parser_mtase_linker = subparsers.add_parser(
        'MTase-linker', 
        help="Commands related to MTase-linker"
    )

    mtase_linker_subparsers = parser_mtase_linker.add_subparsers(
        title="MTase-linker commands", 
        dest="mtase_linker_command"
    )

    parser_mtase_linker_run = mtase_linker_subparsers.add_parser(
        'run', 
        help="Run the MTase-linker workflow"
    )
    parser_mtase_linker_run.add_argument("-t", "--threads", type=int, default=1, help="Number of threads to use. Default is 1")
    parser_mtase_linker_run.add_argument("--forceall", type=bool, default=False, help="Flag for snakemake. Forcerun workflow regardless of already created output (default = False)")
    parser_mtase_linker_run.add_argument("--dryrun", type=bool, default=False, help="Flag for snakemake. Dry-run the workflow. Default is False")
    parser_mtase_linker_run.add_argument("--assembly", type=str, required=True, help="Path to assembly file.")
    parser_mtase_linker_run.add_argument("--contig_bin", type=str, required=True, help="tsv file specifying which bin contigs belong.")
    parser_mtase_linker_run.add_argument("--bin_motifs", type=str, required=True, help="bin-motifs.tsv output from nanomotif.")
    parser_mtase_linker_run.add_argument("-d", "--dependency_dir", type=str, default=os.path.join(os.getcwd(), "ML_dependencies"), help="Path to the ML_dependencies directory created during installation of the MTase-linker module. Default is cwd/ML_dependencies")
    parser_mtase_linker_run.add_argument("-o", "--out", type=str, default=os.path.join(os.getcwd(), "mtase_linker"), help="Path to output directory. Default is cwd")
    parser_mtase_linker_run.add_argument("--identity", type=str, default=80, help="Identity threshold for motif prediction. Default is 80")
    parser_mtase_linker_run.add_argument("--qcovs", type=str, default=80, help="Query coverage for motif prediction. Default is 80")

    parser_mtase_linker_install = mtase_linker_subparsers.add_parser(
        'install', 
        help="Install additional dependencies for MTase-linker"
    )
    parser_mtase_linker_install.add_argument("-d", "--dependency_dir", type=str, default=os.getcwd(), help="Path to installation location of dependencies. A folder named ML_dependencies will be generated. Default is cwd.")
    


    ###########################################################################
    # Check installation
    parser_check_installation = subparsers.add_parser('check_installation', parents=[parser_optional, parser_shared_find_motifs, parser_shared_bin_consensus], add_help=False, help="Performs small test run to verify that the installation is correct.")
    
    
    return parser
    



from setuptools import setup, find_packages
from setuptools.command.install import install
import os
import shutil
import urllib.request

exec(open('nanomotif/_version.py').read())
setup(
    name='nanomotif',
    version=__version__,
    description='Identifying methlyation motifs in nanopore data',
    author='AAU_DarkScience',
    author_email='shei@bio.aau.com',
    license='MIT',
    packages=find_packages(),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    package_data={'nanomotif': [
        'datasets/*',
        'mtase_linker/*.smk',
        '*.yaml',
        "mtase_linker/envs/*",
        "mtase_linker/src/*",
        "bin/methylation_utils"
    ]},
    zip_safe=False,
    install_requires=[
        "wheel",
        "requests",
        "numpy>=1.24.4",
        "pandas>=2.0.2",
        "polars>=0.19,<=0.20.23",
        "scipy>=1.10.1",
        "scikit-learn>=1.5.2",
        "networkx>=3.1",
        "pyarrow>=15.0.2",
        "pymethylation_utils==0.5.2",
        "hdbscan",
        "Bio>=1.6.2",
        "snakemake>=7.32.4",
        "progressbar2>=3.53.1"
    ],
    entry_points={
            'console_scripts': [
                  'nanomotif = nanomotif.main:main'
            ]
    }
)



import polars as pl
from polars import col
import nanomotif as nm
import logging as log
import numpy as np

def within_bin_motifs_consensus(pileup, assembly, motifs, motifs_scored, bins, minimum_contig_motif_methylation=0.25, minimum_bin_methylation=0.75):
    log.debug("Starting within bin motif consensus")

    motifs = motifs.with_columns(pl.lit(True).alias("directly_detected"))
    def convert_motifs_to_regex(motifs):
        return motifs.with_columns(pl.col("motif").map_elements(lambda x: nm.seq.regex_to_iupac(x)).map_elements(lambda x: nm.seq.iupac_to_regex(x)))
    motifs = convert_motifs_to_regex(motifs)
    motifs_scored = convert_motifs_to_regex(motifs_scored)

    motifs_scored = motifs_scored \
        .join(bins, on="contig", how = "left") \
        .filter(pl.col("bin").is_not_null())
    motifs_scored = motifs_scored.join(
        motifs.select("contig", "motif", "mod_type", "mod_position", "directly_detected"), on=["contig", "motif", "mod_type", "mod_position"], how="left"
    ).with_columns(
        pl.col("directly_detected").fill_null(False)
    ).groupby("bin", "motif", "mod_position", "mod_type") \
            .apply(lambda group: group.filter(pl.col("directly_detected").any()))
    
    methylated_contig_mass = motifs_scored.filter(
            ((pl.col("n_mod") / (pl.col("n_mod") + pl.col("n_nomod"))) > minimum_contig_motif_methylation)
        ).groupby(["bin", "motif", "mod_type", "mod_position"]).agg([
            (pl.col("n_mod") + pl.col("n_nomod")).sum().alias("n_motif_meth_contigs")
        ])
    all_contig_mass = motifs_scored.groupby(["bin", "motif", "mod_type", "mod_position"]).agg([
            (pl.col("n_mod") + pl.col("n_nomod")).sum().alias("n_motif_contigs"),
        ])
    result = methylated_contig_mass.join(all_contig_mass, on=["bin", "motif", "mod_type", "mod_position"]) \
        .with_columns(
            (pl.col("n_motif_meth_contigs") / pl.col("n_motif_contigs")).alias("methylated_proportion")
        )
    motifs_scored_filt = motifs_scored.join(result, on=["bin", "motif", "mod_type", "mod_position"]).filter(pl.col("methylated_proportion") > minimum_bin_methylation)

    
    #log.debug("Removing submotifs")
    # Remove submotifs
    #contig_motifs = nm.postprocess.remove_sub_motifs(motifs_scored_filt)

    contig_motifs = motifs_scored_filt.with_columns(
            pl.col("motif").map_elements(lambda x: nm.seq.regex_to_iupac(x), return_dtype = pl.Utf8).alias("motif"),
            (pl.col("n_mod")  / (pl.col("n_mod") + pl.col("n_nomod"))).alias("mean")
        )
    bin_motifs = contig_motifs.groupby("bin", "motif", "mod_position", "mod_type") \
        .agg(
            pl.col("n_mod").sum().alias("n_mod_bin"),
            pl.col("n_nomod").sum().alias("n_nomod_bin"),
            pl.col("contig").count().alias("contig_count")
        ).with_columns( 
            pl.col("motif").apply(lambda x: nm.utils.motif_type(x)).alias("motif_type"),
            (pl.col("n_mod_bin") / (pl.col("n_mod_bin") + pl.col("n_nomod_bin"))).alias("mean_methylation")
        )
    return bin_motifs

def merge_bin_motifs(bin_motifs, bins, pileup, assembly):
    assert list(bin_motifs.schema.keys()) == ['bin', 'motif', 'mod_position', 'mod_type', 'n_mod_bin', 'n_nomod_bin', 'contig_count', 'motif_type', 'mean_methylation']

    for (bin, mod_type), df in bin_motifs.groupby("bin", "mod_type"):
        log.debug(f"Starting motif merge for bin {bin} and mod_type {mod_type}")
        contig_count = df.get_column("contig_count").max()
        
        # Get list of motifs
        motif_seq = df["motif"].to_list()
        motif_pos = df["mod_position"].to_list()
        motifs = [nm.candidate.Motif(nm.candidate.iupac_to_regex(seq), pos) for seq, pos in zip(motif_seq, motif_pos)]
        new_bin_motifs = bin_motifs
        merged_motifs = nm.candidate.merge_motifs(motifs)
        for cluster, motifs in merged_motifs.items():
            log.debug(f"Starting merge for cluster {cluster}")
            merged_motif = motifs[0]
            premerge_motifs = motifs[1]

            premerge_motifs_iupac = [motif.iupac() for motif in premerge_motifs]
            previous_motif_mean_max = df.filter(pl.col("motif").is_in(premerge_motifs_iupac)).get_column("mean_methylation").max()

            merge_motif_n_mod = 0
            merge_motif_n_nomod = 0
            for contig in bins.filter(pl.col("bin") == bin).get_column("contig").unique():
                merge_motif_model = nm.evaluate.motif_model_contig(
                            pileup.pileup.filter((pl.col("contig") == contig) & (pl.col("mod_type") == mod_type)), 
                            assembly.assembly[contig].sequence,
                            merged_motif
                        )
                
                merge_motif_n_mod += merge_motif_model._alpha
                merge_motif_n_nomod += merge_motif_model._beta
            merge_motif_mean = merge_motif_n_mod / (merge_motif_n_mod + merge_motif_n_nomod)

            if merge_motif_mean - previous_motif_mean_max > -0.1:
                log.debug(f"Removing motifs {premerge_motifs_iupac} from bin {bin}")
                new_bin_motifs = new_bin_motifs.filter(
                    pl.col("motif").is_in(premerge_motifs_iupac).not_() | (pl.col("bin") != bin)
                )
                log.debug(f"Adding motif {merged_motif.iupac()} to bin {bin}")
                new_bin_motifs = pl.concat(
                    [new_bin_motifs,
                    pl.DataFrame(
                        {
                            "bin": bin,
                            "motif": merged_motif.iupac(),
                            "mod_position": merged_motif.mod_position,
                            "mod_type": mod_type,
                            "n_mod_bin": merge_motif_n_mod,
                            "n_nomod_bin": merge_motif_n_nomod,
                            "contig_count": contig_count,
                            "motif_type": nm.utils.motif_type(merged_motif.iupac()),
                            "mean_methylation": merge_motif_mean
                        },
                        schema=new_bin_motifs.schema
                    )]
                )
            else:
                log.debug(f"Skipping motif {merged_motif.iupac()} for bin {bin} and mod_type {mod_type} as it has a lower mean methylation than the previous motifs")
    return new_bin_motifs


import polars as pl
import multiprocessing
import os
import warnings
from multiprocessing import get_context
import logging as log
import nanomotif as nm
from nanomotif import candidate, evaluate
from nanomotif.parallel import update_progress_bar
from nanomotif.logger import configure_logger
from nanomotif.utils import concatenate_motif_position_files, clean_up_motif_positions_files
import numpy as np
def score_contig_with_motifs(contig, mod_type, pileup, sequence, motifs, na_positions=None, save_motif_positions=False, positions_outdir=None):

    log.info(f"Scoring motifs in contig {contig}, modtype {mod_type}")
    # Ensuring motifs are in regex for searching
    motifs = motifs\
        .with_columns(motif_iupac = pl.col("motif"))\
        .with_columns([
            pl.col("motif") \
                .map_elements(lambda x: nm.seq.regex_to_iupac(x)) \
                .map_elements(lambda x: nm.seq.iupac_to_regex(x)) \
                .alias("motif")
        ])
    
    # Getting motifs for scoring
    motifs = motifs.select("motif", "motif_iupac", "mod_position", "mod_type").unique()
    motifs = motifs.filter(pl.col("mod_type") == mod_type)    
    motifs_str = motifs.get_column("motif").to_list()
    positions = motifs.get_column("mod_position").to_list()
    types = motifs.get_column("mod_type").to_list()
    motifs_iupac = motifs.get_column("motif_iupac").to_list()

    # Scoring motifs
    models = []
    all_motif_data = {}
    for motif, position, mod_type, motif_iupac in zip(motifs_str, positions, types, motifs_iupac):
        log.debug(f"Scoring motif {motif}, {position}")
        if save_motif_positions:
            model, motif_data =evaluate.motif_model_contig(
                pileup,
                sequence, 
                candidate.Motif(motif, position),
                na_positions=na_positions,
                save_motif_positions=save_motif_positions
            )
            models.append(model)
            all_motif_data[f"{motif_iupac}_{position}_{mod_type}"] = motif_data
        else:
            model = evaluate.motif_model_contig(
                pileup,
                sequence, 
                candidate.Motif(motif, position),
                na_positions=na_positions
            )
            models.append(model)
    
    # Saving all motifs' positions in one file per contig if required
    if save_motif_positions:
        try:
            path = f'{positions_outdir}/{contig}_{mod_type}_motifs_positions.npz'
            np.savez(path, **all_motif_data)
        except Exception as e:
            log.error(f"Failed to save motif positions for contig {contig}: {e}")
            
    # Creating a dataframe of the results
    result = pl.DataFrame({
            "motif": motifs_str,
            "mod_position": positions,
            "mod_type": types,
            "model": models
        })
    result = result.with_columns([
            pl.col("model").map_elements(lambda x: x._alpha).alias("n_mod"),
            pl.col("model").map_elements(lambda x: x._beta).alias("n_nomod"),
            pl.col("motif").map_elements(lambda x: nm.seq.regex_to_iupac(x)).alias("motif"),
            pl.lit(contig).alias("contig")
        ]).drop("model")
    return result

def worker_function(args, counter, lock, sequence, motifs, log_dir, verbose, seed, save_motif_positions = False, positions_outdir = None):
    """
    Process a single subpileup for one contig and one modtype

    Parameters:
    - args (tuple): The arguments to the function: contig, modtype, subpileup
    - counter (multiprocessing.Value): The progress counter
    - lock (multiprocessing.Lock): The lock for the progress counter
    - assembly (Assembly): The assembly to be processed.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - padding (int): The padding to use for the motif.
    """
    warnings.filterwarnings("ignore")
    nm.seed.set_seed(seed)
    contig, modtype, subpileup, na_positions = args

    if na_positions is not None:
        na_positions = na_positions.explode("position", "strand")
        na_positions = {
            "fwd": na_positions.filter(pl.col("strand") == "+")["position"].to_numpy(),
            "rev": na_positions.filter(pl.col("strand") == "-")["position"].to_numpy()
        }
    
    process_id = os.getpid()
    log_file = f"{log_dir}/score-motifs.{process_id}.log"
    configure_logger(log_file, verbose=verbose)
    try:
        result = score_contig_with_motifs(contig, modtype, subpileup, sequence, motifs, save_motif_positions=save_motif_positions, positions_outdir=positions_outdir, na_positions=na_positions)
        with lock:
            counter.value += 1
        return result
    except:
        with lock:
            counter.value += 1
        return None

def score_sample_parallel(
        assembly, pileup, motifs,
        na_position = None,
        threads = 2,
        threshold_methylation_general = 0.5,
        threshold_valid_coverage = 1,
        verbose = False,
        log_dir = None,
        save_motif_positions = False,
        positions_outdir = None,
        seed = None
    ):
    """
    Process a single sample
    
    Parameters:
    - assembly (Assembly): The assembly to be processed.
    - pileup (Pileup): The pileup to be processed.
    - max_candidate_size (int): The maximum size of the candidate motifs.
    - min_read_methylation_fraction (float): The minimum fraction of reads that must be methylated for a position to be considered methylated.
    - threshold_valid_coverage (int): The minimum number of reads that must cover a position for it to be considered valid.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - min_cdf_score (float): Minimum score of 1 - cdf(cdf_position) for a motif to be considered valid.
    - cdf_position (float): The position to evaluate the cdf at.
    - min_motif_frequency (int): Used to get minimum number of sequences to evaluate motif at.
    - save_motif_positions (bool): Save motif positions in the output folder.
    - positions_outdir (str): The output directory for motif positions.
    """
    assert pileup is not None, "Pileup is None"
    assert assembly is not None, "Assembly is None"
    assert threshold_methylation_general >= 0 and threshold_methylation_general <= 1, "threshold_methylation_general must be between 0 and 1"
    assert threshold_valid_coverage >= 0, "min_valid_coverage must be greater than 0"

    # Filter pileup
    pileup = pileup \
            .filter(pl.col("Nvalid_cov") > threshold_valid_coverage) \
            .filter(pl.col("fraction_mod") >= threshold_methylation_general) \
            .sort("contig") 
    
    # Create a list of tasks (TODO: not have a list of all data)
    if na_position is not None:
        tasks = [(contig, modtype, subpileup, na_position.filter((pl.col("contig")==contig) & (pl.col("mod_type") == modtype))) for (contig, modtype), subpileup in pileup.groupby(["contig", "mod_type"])]
    else:
        tasks = [(contig, modtype, subpileup, None) for (contig, modtype), subpileup in pileup.groupby(["contig", "mod_type"])]
    # Create a progress manager
    manager = multiprocessing.Manager()
    counter = manager.Value('i', 0)
    lock = manager.Lock()

    # Create a pool of workers
    pool = get_context("spawn").Pool(processes=threads)

    # Create a process for the progress bar
    progress_bar_process = multiprocessing.Process(target=update_progress_bar, args=(counter, len(tasks), True))
    progress_bar_process.start()

    # Put them workers to work
    results = pool.starmap(worker_function, [(
        task, 
        counter, lock, 
        assembly.assembly[task[0]].sequence, motifs,
        log_dir, verbose, seed, save_motif_positions, positions_outdir
        ) for task in tasks])
    results = [result for result in results if result is not None]

    # Close the pool
    pool.close()
    pool.join()

    # Close the progress bar
    progress_bar_process.join()

    if len(results) == 0:
        return None
    motifs_all_scored = pl.concat(results)
    
    if save_motif_positions:
        files = [os.path.join(positions_outdir, f) for f in os.listdir(positions_outdir) if f.endswith("_motifs_positions.npz")]
        concatenate_motif_position_files(files, positions_outdir)
        clean_up_motif_positions_files(files)
    
    return motifs_all_scored



def score_all_contigs_all_motifs(pileup, assm, motifs, frac_mod=0.7):
    # Score all contigs with identified motifs
    all_motifs_scored = {
        "contig":[],
        "type":[],
        "motif":[],
        "mod_position":[],
        "model":[],
        "mean":[]
    }
    for contig in pileup.pileup.get_column("contig").unique().to_list():
        for row in motifs.select("motif", "mod_position", "mod_type").unique().rows():
            model = evaluate.motif_model_contig(
                pileup.pileup.filter(pl.col("Nvalid_cov") > 1).filter(pl.col("fraction_mod") > frac_mod),
                assm.assembly[contig].sequence, 
                candidate.Motif(row[0], row[1])
            )
            all_motifs_scored["contig"].append(contig)
            all_motifs_scored["motif"].append(row[0])
            all_motifs_scored["model"].append(model)
            all_motifs_scored["mean"].append(model.mean())
            all_motifs_scored["type"].append(row[2])
            all_motifs_scored["mod_position"].append(row[1])

    scored_all = pl.DataFrame(all_motifs_scored).with_columns([
        pl.col("model").map_elements(lambda x: x._alpha).alias("alpha"),
        pl.col("model").map_elements(lambda x: x._beta).alias("beta")
    ])
    return scored_all
